"""SEC Filings fetcher."""


from typing import Any, Dict, List, Optional

from openbb_provider.abstract.fetcher import Fetcher
from openbb_provider.standard_models.sec_filings import (
    SECFilingsData,
    SECFilingsQueryParams,
)

from openbb_fmp.utils.helpers import create_url, get_data_many


class FMPSECFilingsQueryParams(SECFilingsQueryParams):
    """FMP SEC Filings Query.

    Source: https://site.financialmodelingprep.com/developer/docs/sec-filings-api/
    """


class FMPSECFilingsData(SECFilingsData):
    """FMP SEC Filings Data."""


class FMPSECFilingsFetcher(
    Fetcher[
        FMPSECFilingsQueryParams,
        List[FMPSECFilingsData],
    ]
):
    """Transform the query, extract and transform the data from the FMP endpoints."""

    @staticmethod
    def transform_query(params: Dict[str, Any]) -> FMPSECFilingsQueryParams:
        """Transform the query params."""
        return FMPSECFilingsQueryParams(**params)

    @staticmethod
    def extract_data(
        query: FMPSECFilingsQueryParams,
        credentials: Optional[Dict[str, str]],
        **kwargs: Any,
    ) -> List[Dict]:
        """Return the raw data from the FMP endpoint."""
        api_key = credentials.get("fmp_api_key") if credentials else ""

        url = create_url(
            3, f"sec_filings/{query.symbol}", api_key, query, exclude=["symbol"]
        )

        return get_data_many(url, **kwargs)

    @staticmethod
    def transform_data(data: List[Dict]) -> List[FMPSECFilingsData]:
        """Return the transformed data."""
        return [FMPSECFilingsData(**d) for d in data]



"""FMP Stock Peers fetcher."""


from typing import Any, Dict, Optional

from openbb_provider.abstract.fetcher import Fetcher
from openbb_provider.standard_models.stock_peers import (
    StockPeersData,
    StockPeersQueryParams,
)

from openbb_fmp.utils.helpers import create_url, get_data_one

# FMP SPECIFIC FUNCTIONALITY CURRENTLY


class FMPStockPeersQueryParams(StockPeersQueryParams):
    """FMP Stock Peers query.

    Source: https://site.financialmodelingprep.com/developer/docs/#Stock-Peers
    """


class FMPStockPeersData(StockPeersData):
    """FMP Stock Peers data."""


class FMPStockPeersFetcher(
    Fetcher[
        FMPStockPeersQueryParams,
        FMPStockPeersData,
    ]
):
    """Transform the query, extract and transform the data from the FMP endpoints."""

    @staticmethod
    def transform_query(params: Dict[str, Any]) -> FMPStockPeersQueryParams:
        """Transform the query params."""
        return FMPStockPeersQueryParams(**params)

    @staticmethod
    def extract_data(
        query: FMPStockPeersQueryParams,
        credentials: Optional[Dict[str, str]],
        **kwargs: Any,
    ) -> Dict:
        """Return the raw data from the FMP endpoint."""
        api_key = credentials.get("fmp_api_key") if credentials else ""
        url = create_url(4, "stock_peers", api_key, query)

        return get_data_one(url, **kwargs)

    @staticmethod
    def transform_data(data: Dict) -> FMPStockPeersData:
        """Return the transformed data."""
        return FMPStockPeersData(**data)



"""FMP Income Statement Fetcher."""


from datetime import (
    date as dateType,
    datetime,
)
from typing import Any, Dict, List, Literal, Optional

from openbb_provider.abstract.fetcher import Fetcher
from openbb_provider.standard_models.income_statement import (
    IncomeStatementData,
    IncomeStatementQueryParams,
)
from pydantic import Field, root_validator

from openbb_fmp.utils.helpers import get_data_many

PeriodType = Literal["annual", "quarter"]


class FMPIncomeStatementQueryParams(IncomeStatementQueryParams):
    """FMP Income Statement QueryParams.

    Source: https://financialmodelingprep.com/developer/docs/#Income-Statement
    """

    symbol: str = Field(description="Symbol/CIK of the company.")

    @root_validator()
    def check_symbol_or_cik(cls, values):  # pylint: disable=no-self-argument
        """Validate that either a symbol or CIK is provided."""
        if values.get("symbol") is None and values.get("cik") is None:
            raise ValueError("symbol or cik must be provided")
        return values


class FMPIncomeStatementData(IncomeStatementData):
    """FMP Income Statement Data."""

    class Config:
        """Pydantic alias config using fields dict."""

        fields = {
            "cost_of_revenue": "costOfRevenue",
            "gross_profit": "grossProfit",
            "gross_profit_ratio": "grossProfitRatio",
            "research_and_development_expenses": "researchAndDevelopmentExpenses",
            "general_and_administrative_expenses": "generalAndAdministrativeExpenses",
            "selling_and_marketing_expenses": "sellingAndMarketingExpenses",
            "selling_general_and_administrative_expenses": "sellingGeneralAndAdministrativeExpenses",
            "other_expenses": "otherExpenses",
            "operating_expenses": "operatingExpenses",
            "depreciation_and_amortization": "depreciationAndAmortization",
            "ebitda_ratio": "ebitdaratio",
            "operating_income": "operatingIncome",
            "operating_income_ratio": "operatingIncomeRatio",
            "interest_income": "interestIncome",
            "interest_expense": "interestExpense",
            "total_other_income_expenses_net": "totalOtherIncomeExpensesNet",
            "income_before_tax": "incomeBeforeTax",
            "income_before_tax_ratio": "incomeBeforeTaxRatio",
            "income_tax_expense": "incomeTaxExpense",
            "net_income": "netIncome",
            "net_income_ratio": "netIncomeRatio",
            "eps_diluted": "epsdiluted",
            "weighted_average_shares_outstanding": "weightedAverageShsOut",
            "weighted_average_shares_outstanding_dil": "weightedAverageShsOutDil",
        }

        reported_currency: Optional[str] = Field(description="Reporting currency.")
        filing_date: Optional[dateType] = Field(description="Filling date.")
        accepted_date: Optional[datetime] = Field(description="Accepted date.")
        calendar_year: Optional[int] = Field(description="Calendar year.")

        cost_of_revenue: Optional[float] = Field(description="Cost of revenue.")
        gross_profit: Optional[float] = Field(description="Gross profit.")
        gross_profit_ratio: Optional[float] = Field(description="Gross profit ratio.")
        research_and_development_expenses: Optional[float] = Field(
            description="Research and development expenses."
        )
        general_and_administrative_expenses: Optional[float] = Field(
            description="General and administrative expenses."
        )
        selling_and_marketing_expenses: Optional[float] = Field(
            description="Selling and marketing expenses."
        )
        selling_general_and_administrative_expenses: Optional[float] = Field(
            description="Selling, general and administrative expenses."
        )
        other_expenses: Optional[float] = Field(description="Other expenses.")

        operating_expenses: Optional[float] = Field(description="Operating expenses.")
        depreciation_and_amortization: Optional[float] = Field(
            description="Depreciation and amortization."
        )
        ebitda_ratio: Optional[float] = Field(description="EBIDTA ratio.")
        operating_income: Optional[float] = Field(description="Operating income.")
        operating_income_ratio: Optional[float] = Field(
            description="Operating income ratio."
        )
        interest_income: Optional[float] = Field(description="Interest income.")
        interest_expense: Optional[float] = Field(description="Interest expense.")

        total_other_income_expenses_net: Optional[float] = Field(
            description="Total other income expenses net."
        )
        income_before_tax: Optional[float] = Field(description="Income before tax.")
        income_before_tax_ratio: Optional[float] = Field(
            description="Income before tax ratio."
        )
        income_tax_expense: Optional[float] = Field(description="Income tax expense.")
        net_income: Optional[float] = Field(description="Net income.")
        net_income_ratio: Optional[float] = Field(description="Net income ratio.")

        link: Optional[str] = Field(description="Link to the income statement.")
        final_link: Optional[str] = Field(
            description="Final link to the income statement."
        )


class FMPIncomeStatementFetcher(
    Fetcher[
        FMPIncomeStatementQueryParams,
        List[FMPIncomeStatementData],
    ]
):
    """Transform the query, extract and transform the data from the FMP endpoints."""

    @staticmethod
    def transform_query(params: Dict[str, Any]) -> FMPIncomeStatementQueryParams:
        """Transform the query params."""
        return FMPIncomeStatementQueryParams(**params)

    @staticmethod
    def extract_data(
        query: FMPIncomeStatementQueryParams,
        credentials: Optional[Dict[str, str]],
        **kwargs: Any,
    ) -> List[Dict]:
        """Return the raw data from the FMP endpoint."""
        api_key = credentials.get("fmp_api_key") if credentials else ""

        symbol = query.symbol or query.cik
        base_url = "https://financialmodelingprep.com/api/v3"

        url = (
            f"{base_url}/income-statement/{symbol}?"
            f"period={query.period}&limit={query.limit}&apikey={api_key}"
        )

        return get_data_many(url, **kwargs)

    @staticmethod
    def transform_data(data: List[Dict]) -> List[FMPIncomeStatementData]:
        """Return the transformed data."""
        return [FMPIncomeStatementData(**d) for d in data]



"""Key Metrics fetcher."""


from concurrent.futures import ThreadPoolExecutor
from itertools import repeat
from typing import Any, Dict, List, Optional

from openbb_provider.abstract.fetcher import Fetcher
from openbb_provider.standard_models.key_metrics import (
    KeyMetricsData,
    KeyMetricsQueryParams,
)

from openbb_fmp.utils.helpers import create_url, get_data_many


class FMPKeyMetricsQueryParams(KeyMetricsQueryParams):
    """FMP Key Metrics Query.

    Source: https://site.financialmodelingprep.com/developer/docs/company-key-metrics-api/
    """


class FMPKeyMetricsData(KeyMetricsData):
    """FMP Key Metrics Data."""

    class Config:
        """Pydantic alias config using fields dict."""

        fields = {
            "pocf_ratio": "pocfratio",
            "research_and_development_to_revenue": "researchAndDdevelopementToRevenue",
        }


class FMPKeyMetricsFetcher(
    Fetcher[
        FMPKeyMetricsQueryParams,
        List[FMPKeyMetricsData],
    ]
):
    """Transform the query, extract and transform the data from the FMP endpoints."""

    @staticmethod
    def transform_query(params: Dict[str, Any]) -> FMPKeyMetricsQueryParams:
        """Transform the query params."""
        return FMPKeyMetricsQueryParams(**params)

    @staticmethod
    def extract_data(
        query: FMPKeyMetricsQueryParams,
        credentials: Optional[Dict[str, str]],
        **kwargs: Any,
    ) -> List[Dict]:
        """Return the raw data from the FMP endpoint."""
        api_key = credentials.get("fmp_api_key") if credentials else ""

        data: List[Dict] = []

        def multiple_symbols(symbol: str, data: List[Dict]) -> None:
            url = create_url(
                3, f"key-metrics/{symbol}", api_key, query, exclude=["symbol"]
            )
            return data.extend(get_data_many(url, **kwargs))

        with ThreadPoolExecutor() as executor:
            executor.map(multiple_symbols, query.symbol.split(","), repeat(data))

        return data

    @staticmethod
    def transform_data(data: List[Dict]) -> List[FMPKeyMetricsData]:
        """Return the transformed data."""
        return [FMPKeyMetricsData(**d) for d in data]



"""FMP Stock News fetcher."""


from typing import Any, Dict, List, Optional

from openbb_provider.abstract.fetcher import Fetcher
from openbb_provider.standard_models.stock_news import (
    StockNewsData,
    StockNewsQueryParams,
)
from pydantic import Field

from openbb_fmp.utils.helpers import create_url, get_data_many


class FMPStockNewsQueryParams(StockNewsQueryParams):
    """FMP Stock News query.

    Source: https://site.financialmodelingprep.com/developer/docs/stock-news-api/
    """

    class Config:
        """Pydantic alias config using fields dict."""

        fields = {"symbols": "tickers"}


class FMPStockNewsData(StockNewsData):
    """FMP Stock News data."""

    class Config:
        """Pydantic alias config using fields dict."""

        fields = {"date": "publishedDate"}

    symbol: str = Field(description="Ticker of the fetched news.")
    image: Optional[str] = Field(description="URL to the image of the news source.")
    site: str = Field(description="Name of the news source.")


class FMPStockNewsFetcher(
    Fetcher[
        FMPStockNewsQueryParams,
        List[FMPStockNewsData],
    ]
):
    """Transform the query, extract and transform the data from the FMP endpoints."""

    @staticmethod
    def transform_query(params: Dict[str, Any]) -> FMPStockNewsQueryParams:
        """Transform the query params."""
        return FMPStockNewsQueryParams(**params)

    @staticmethod
    def extract_data(
        query: FMPStockNewsQueryParams,
        credentials: Optional[Dict[str, str]],
        **kwargs: Any,
    ) -> List[Dict]:
        """Return the raw data from the FMP endpoint."""
        api_key = credentials.get("fmp_api_key") if credentials else ""
        url = create_url(3, "stock_news", api_key, query)

        return get_data_many(url, **kwargs)

    @staticmethod
    def transform_data(data: List[Dict]) -> List[FMPStockNewsData]:
        """Return the transformed data."""
        return [FMPStockNewsData(**d) for d in data]



"""FMP Stock Multiples Fetcher."""


from typing import Any, Dict, List, Optional

from openbb_provider.abstract.fetcher import Fetcher
from openbb_provider.standard_models.stock_multiples import (
    StockMultiplesData,
    StockMultiplesQueryParams,
)

from openbb_fmp.utils.helpers import create_url, get_data_many


class FMPStockMultiplesQueryParams(StockMultiplesQueryParams):
    """FMP Stock Multiples Query.

    Source: https://site.financialmodelingprep.com/developer/docs/#Company-Key-Metrics
    """


class FMPStockMultiplesData(StockMultiplesData):
    """FMP Stock Multiples Data."""

    class Config:
        """Pydantic alias config using fields dict."""

        fields = {
            "pocf_ratio_ttm": "pocfratioTTM",
            "enterprise_value_over_ebitda_ttm": "enterpriseValueOverEBITDATTM",
            "net_debt_to_ebitda_ttm": "netDebtToEBITDATTM",
            "research_and_development_to_revenue_ttm": "researchAndDevelopementToRevenueTTM",
        }


class FMPStockMultiplesFetcher(
    Fetcher[
        FMPStockMultiplesQueryParams,
        List[FMPStockMultiplesData],
    ]
):
    """Transform the query, extract and transform the data from the FMP endpoints."""

    @staticmethod
    def transform_query(params: Dict[str, Any]) -> FMPStockMultiplesQueryParams:
        """Transform the query params."""
        return FMPStockMultiplesQueryParams(**params)

    @staticmethod
    def extract_data(
        query: FMPStockMultiplesQueryParams,
        credentials: Optional[Dict[str, str]],
        **kwargs: Any,
    ) -> List[Dict]:
        """Return the raw data from the FMP endpoint."""
        api_key = credentials.get("fmp_api_key") if credentials else ""
        url = create_url(
            3, f"key-metrics-ttm/{query.symbol}", api_key, query, exclude=["symbol"]
        )

        return get_data_many(url, **kwargs)

    @staticmethod
    def transform_data(data: List[Dict]) -> List[FMPStockMultiplesData]:
        """Return the transformed data."""
        return [FMPStockMultiplesData(**d) for d in data]



"""FMP Major Indices end of day fetcher."""


from datetime import datetime
from typing import Any, Dict, List, Literal, Optional

from openbb_provider.abstract.fetcher import Fetcher
from openbb_provider.standard_models.major_indices_historical import (
    MajorIndicesHistoricalData,
    MajorIndicesHistoricalQueryParams,
)
from openbb_provider.utils.helpers import get_querystring
from pydantic import Field, NonNegativeInt, validator

from openbb_fmp.utils.helpers import get_data_many


class FMPMajorIndicesHistoricalQueryParams(MajorIndicesHistoricalQueryParams):
    """FMP Major Indices end of day Query.

    Source: https://site.financialmodelingprep.com/developer/docs/historical-index-price-api/
    """

    timeseries: Optional[NonNegativeInt] = Field(
        default=None, description="Number of days to look back."
    )
    interval: Literal[
        "1min", "5min", "15min", "30min", "1hour", "4hour", "1day"
    ] = Field(default="1day", description="Interval of the data to fetch.")


class FMPMajorIndicesHistoricalData(MajorIndicesHistoricalData):
    """FMP Major Indices end of day Data."""

    adj_close: Optional[float] = Field(
        description="Adjusted Close Price of the symbol.",
        default=None,
    )
    unadjusted_volume: Optional[float] = Field(
        description="Unadjusted volume of the symbol.",
        default=None,
    )
    change: Optional[float] = Field(
        description="Change in the price of the symbol from the previous day.",
        default=None,
    )
    change_percent: Optional[float] = Field(
        description=r"Change \% in the price of the symbol.",
        default=None,
    )
    label: Optional[str] = Field(
        description="Human readable format of the date.", default=None
    )
    change_over_time: Optional[float] = Field(
        description=r"Change \% in the price of the symbol over a period of time.",
        default=None,
    )

    @validator("date", pre=True)
    def date_validate(cls, v) -> datetime:  # pylint: disable=E0213
        """Return the date as a datetime object."""
        # TODO: Check if this condition is necessary
        # if values.get("changeOverTime", None) is not None:
        #     return datetime.strptime(v, "%Y-%m-%d")
        return datetime.strptime(v, "%Y-%m-%d")


class FMPMajorIndicesHistoricalFetcher(
    Fetcher[
        FMPMajorIndicesHistoricalQueryParams,
        List[FMPMajorIndicesHistoricalData],
    ]
):
    """Transform the query, extract and transform the data from the FMP endpoints."""

    @staticmethod
    def transform_query(params: Dict[str, Any]) -> FMPMajorIndicesHistoricalQueryParams:
        """Transform the query params."""
        return FMPMajorIndicesHistoricalQueryParams(**params)

    @staticmethod
    def extract_data(
        query: FMPMajorIndicesHistoricalQueryParams,
        credentials: Optional[Dict[str, str]],
        **kwargs: Any,
    ) -> List[Dict]:
        """Return the raw data from the FMP endpoint."""
        api_key = credentials.get("fmp_api_key") if credentials else ""

        base_url = "https://financialmodelingprep.com/api/v3"
        query_str = (
            get_querystring(query.dict(), ["symbol"])
            .replace("start_date", "from")
            .replace("end_date", "to")
        )

        url_params = f"{query.symbol}?{query_str}&apikey={api_key}"
        url = f"{base_url}/historical-chart/{query.interval}/%5E{url_params}"

        if query.interval == "1day":
            url = f"{base_url}/historical-price-full/index/%5E{url_params}"

        return get_data_many(url, "historical", **kwargs)

    @staticmethod
    def transform_data(data: List[Dict]) -> List[FMPMajorIndicesHistoricalData]:
        """Return the transformed data."""
        return [FMPMajorIndicesHistoricalData(**d) for d in data]



"""FMP Balance Sheet Fetcher."""


from datetime import (
    date as dateType,
    datetime,
)
from typing import Any, Dict, List, Optional

from openbb_provider.abstract.fetcher import Fetcher
from openbb_provider.standard_models.balance_sheet import (
    BalanceSheetData,
    BalanceSheetQueryParams,
)
from pydantic import Field, validator

from openbb_fmp.utils.helpers import create_url, get_data_many


class FMPBalanceSheetQueryParams(BalanceSheetQueryParams):
    """FMP Balance Sheet QueryParams.

    Source: https://financialmodelingprep.com/developer/docs/#Balance-Sheet
    """

    symbol: str = Field(description="Symbol/CIK of the company.")


class FMPBalanceSheetData(BalanceSheetData):
    """FMP Balance Sheet Data."""

    class Config:
        """Pydantic alias config using fields Dict."""

        fields = {
            "marketable_securities": "longTermInvestments",
            "other_shareholder_equity": "othertotalStockholdersEquity",
            "total_shareholder_equity": "totalStockholdersEquity",
            "total_liabilities_and_shareholders_equity": "totalLiabilitiesAndStockholdersEquity",
        }

    reported_currency: str = Field(description="Reported currency in the statement.")

    filling_date: dateType = Field(description="Filling date.")
    accepted_date: datetime = Field(description="Accepted date.")
    calendar_year: int = Field(description="Calendar year.")

    cash_and_short_term_investments: Optional[int] = Field(
        description="Cash and short term investments"
    )
    goodwill_and_intangible_assets: Optional[int] = Field(
        description="Goodwill and Intangible Assets"
    )
    capital_lease_obligations: Optional[int] = Field(
        description="Capital lease obligations"
    )
    total_investments: Optional[int] = Field(description="Total investments")
    total_debt: Optional[int] = Field(description="Total debt")
    net_debt: Optional[int] = Field(description="Net debt")

    link: str = Field(description="Link to the statement.")
    final_link: str = Field(description="Link to the final statement.")

    @validator("filing_date", pre=True, check_fields=False)
    def filing_date_validate(cls, v):  # pylint: disable=no-self-argument
        """Validate the filing date."""
        return datetime.strptime(v, "%Y-%m-%d").date()

    @validator("accepted_date", pre=True, check_fields=False)
    def accepted_date_validate(cls, v):  # pylint: disable=no-self-argument
        """Validate the accepted date."""
        return datetime.strptime(v, "%Y-%m-%d %H:%M:%S")


class FMPBalanceSheetFetcher(
    Fetcher[
        FMPBalanceSheetQueryParams,
        List[FMPBalanceSheetData],
    ]
):
    """Transform the query, extract and transform the data from the FMP endpoints."""

    @staticmethod
    def transform_query(params: Dict[str, Any]) -> FMPBalanceSheetQueryParams:
        """Transform the query params."""
        return FMPBalanceSheetQueryParams(**params)

    @staticmethod
    def extract_data(
        query: FMPBalanceSheetQueryParams,
        credentials: Optional[Dict[str, str]],
        **kwargs: Any,
    ) -> List[Dict]:
        """Return the raw data from the FMP endpoint."""
        api_key = credentials.get("fmp_api_key") if credentials else ""

        url = create_url(
            3, f"balance-sheet-statement/{query.symbol}", api_key, query, ["symbol"]
        )

        return get_data_many(url, **kwargs)

    @staticmethod
    def transform_data(data: List[Dict]) -> List[FMPBalanceSheetData]:
        """Return the transformed data."""
        return [FMPBalanceSheetData(**d) for d in data]



"""FMP Stocks end of day fetcher."""


from datetime import datetime
from typing import Any, Dict, List, Optional

from openbb_provider.abstract.fetcher import Fetcher
from openbb_provider.standard_models.stock_quote import (
    StockQuoteData,
    StockQuoteQueryParams,
)
from pydantic import Field, validator

from openbb_fmp.utils.helpers import get_data_many, get_querystring


class FMPStockQuoteQueryParams(StockQuoteQueryParams):
    """FMP Stock end of day Query.

    Source: https://financialmodelingprep.com/developer/docs/#Stock-Historical-Price
    """


class FMPStockQuoteData(StockQuoteData):
    """FMP Stock end of day Data."""

    class Config:
        """Pydantic alias config using fields dict."""

        fields = {
            "date": "timestamp",
        }

    symbol: Optional[str] = Field(description="Symbol of the company.")
    name: Optional[str] = Field(description="Name of the company.")
    price: Optional[float] = Field(description="Current trading price of the stock.")
    changes_percentage: Optional[float] = Field(
        description="Change percentage of the stock price."
    )
    change: Optional[float] = Field(description="Change in the stock price.")
    year_high: Optional[float] = Field(
        description="Highest price of the stock in the last 52 weeks."
    )
    year_low: Optional[float] = Field(
        description="Lowest price of the stock in the last 52 weeks."
    )
    market_cap: Optional[float] = Field(description="Market cap of the company.")
    price_avg50: Optional[float] = Field(
        description="50 days average price of the stock."
    )
    price_avg200: Optional[float] = Field(
        description="200 days average price of the stock."
    )
    volume: Optional[int] = Field(
        description="Volume of the stock in the current trading day."
    )
    avg_volume: Optional[int] = Field(
        default=None,
        description="Average volume of the stock in the last 10 trading days.",
    )
    exchange: Optional[str] = Field(description="Exchange the stock is traded on.")
    open: Optional[float] = Field(
        default=None,
        description="Opening price of the stock in the current trading day.",
    )
    previous_close: Optional[float] = Field(
        description="Previous closing price of the stock."
    )
    eps: Optional[float] = Field(description="Earnings per share of the stock.")
    pe: Optional[float] = Field(description="Price earnings ratio of the stock.")
    earnings_announcement: Optional[str] = Field(
        description="Earnings announcement date of the stock."
    )
    shares_outstanding: Optional[int] = Field(
        description="Number of shares outstanding of the stock."
    )

    @validator("timestamp", pre=True, check_fields=False)
    def date_validate(cls, v):  # pylint: disable=E0213
        """Return the date as a datetime object."""
        return datetime.strptime(v, "%Y-%m-%d")


class FMPStockQuoteFetcher(
    Fetcher[
        FMPStockQuoteQueryParams,
        List[FMPStockQuoteData],
    ]
):
    """Transform the query, extract and transform the data from the FMP endpoints."""

    @staticmethod
    def transform_query(params: Dict[str, Any]) -> FMPStockQuoteQueryParams:
        """Transform the query params."""
        return FMPStockQuoteQueryParams(**params)

    @staticmethod
    def extract_data(
        query: FMPStockQuoteQueryParams,
        credentials: Optional[Dict[str, str]],
        **kwargs: Any,
    ) -> List[Dict]:
        """Return the raw data from the FMP endpoint."""
        api_key = credentials.get("fmp_api_key") if credentials else ""

        base_url = "https://financialmodelingprep.com/api/v3"
        query_str = get_querystring(query.dict(), ["symbol"])
        url = f"{base_url}/quote/{query.symbol}?{query_str}&apikey={api_key}"

        return get_data_many(url, **kwargs)

    @staticmethod
    def transform_data(data: List[Dict]) -> List[FMPStockQuoteData]:
        """Return the transformed data."""
        return [FMPStockQuoteData(**d) for d in data]



"""FMP Global News fetcher."""

from typing import Any, Dict, List, Optional

from openbb_provider.abstract.fetcher import Fetcher
from openbb_provider.standard_models.global_news import (
    GlobalNewsData,
    GlobalNewsQueryParams,
)
from pydantic import Field

from openbb_fmp.utils.helpers import create_url, get_data_many


class FMPGlobalNewsQueryParams(GlobalNewsQueryParams):
    """FMP Global News query.

    Source: https://site.financialmodelingprep.com/developer/docs/general-news-api/
    """


class FMPGlobalNewsData(GlobalNewsData):
    """FMP Global News Data."""

    class Config:
        """Pydantic alias config using fields dict."""

        fields = {"date": "publishedDate"}

    site: str = Field(description="Site of the news.")


class FMPGlobalNewsFetcher(
    Fetcher[
        FMPGlobalNewsQueryParams,
        List[FMPGlobalNewsData],
    ]
):
    """Transform the query, extract and transform the data from the FMP endpoints."""

    @staticmethod
    def transform_query(params: Dict[str, Any]) -> FMPGlobalNewsQueryParams:
        """Transform the query params."""
        return FMPGlobalNewsQueryParams(**params)

    @staticmethod
    def extract_data(
        query: FMPGlobalNewsQueryParams,
        credentials: Optional[Dict[str, str]],
        **kwargs: Any,
    ) -> List[Dict]:
        """Return the raw data from the FMP endpoint."""
        api_key = credentials.get("fmp_api_key") if credentials else ""

        url = create_url(4, "general_news", api_key, query)

        return get_data_many(url, **kwargs)

    @staticmethod
    def transform_data(data: List[Dict]) -> List[FMPGlobalNewsData]:
        """Return the transformed data."""
        return [FMPGlobalNewsData(**d) for d in data]



"""FMP Stock Ownership fetcher."""


from typing import Any, Dict, List, Optional

from openbb_provider.abstract.fetcher import Fetcher
from openbb_provider.standard_models.stock_ownership import (
    StockOwnershipData,
    StockOwnershipQueryParams,
)

from openbb_fmp.utils.helpers import create_url, get_data_many


class FMPStockOwnershipQueryParams(StockOwnershipQueryParams):
    """FMP Stock Ownership query.

    Source: https://site.financialmodelingprep.com/developer/docs/#Stock-Ownership-by-Holders
    """


class FMPStockOwnershipData(StockOwnershipData):
    """FMP Stock Ownership Data."""


class FMPStockOwnershipFetcher(
    Fetcher[
        FMPStockOwnershipQueryParams,
        List[FMPStockOwnershipData],
    ]
):
    """Transform the query, extract and transform the data from the FMP endpoints."""

    @staticmethod
    def transform_query(params: Dict[str, Any]) -> FMPStockOwnershipQueryParams:
        """Transform the query params."""
        return FMPStockOwnershipQueryParams(**params)

    @staticmethod
    def extract_data(
        query: FMPStockOwnershipQueryParams,
        credentials: Optional[Dict[str, str]],
        **kwargs: Any,
    ) -> List[Dict]:
        """Return the raw data from the FMP endpoint."""
        api_key = credentials.get("fmp_api_key") if credentials else ""
        url = create_url(
            4,
            "institutional-ownership/institutional-holders/symbol-ownership-percent",
            api_key,
            query,
        )

        return get_data_many(url, **kwargs)

    @staticmethod
    def transform_data(data: List[Dict]) -> List[FMPStockOwnershipData]:
        """Return the transformed data."""
        own = [FMPStockOwnershipData(**d) for d in data]
        own.sort(key=lambda x: x.filing_date, reverse=True)
        return own



"""FMP Stock Insider Trading fetcher."""


from typing import Any, Dict, List, Optional

from openbb_provider.abstract.fetcher import Fetcher
from openbb_provider.standard_models.stock_insider_trading import (
    StockInsiderTradingData,
    StockInsiderTradingQueryParams,
)

from openbb_fmp.utils.helpers import create_url, get_data_many


class FMPStockInsiderTradingQueryParams(StockInsiderTradingQueryParams):
    """FMP Stock Insider Trading Query.

    Source: https://site.financialmodelingprep.com/developer/docs/#Stock-Insider-Trading
    """


class FMPStockInsiderTradingData(StockInsiderTradingData):
    """FMP Stock Insider Trading Data."""


class FMPStockInsiderTradingFetcher(
    Fetcher[
        FMPStockInsiderTradingQueryParams,
        List[FMPStockInsiderTradingData],
    ]
):
    """Transform the query, extract and transform the data from the FMP endpoints."""

    @staticmethod
    def transform_query(params: Dict[str, Any]) -> FMPStockInsiderTradingQueryParams:
        """Transform the query params."""
        return FMPStockInsiderTradingQueryParams(**params)

    @staticmethod
    def extract_data(
        query: FMPStockInsiderTradingQueryParams,
        credentials: Optional[Dict[str, str]],
        **kwargs: Any,
    ) -> List[Dict]:
        """Return the raw data from the FMP endpoint."""
        api_key = credentials.get("fmp_api_key") if credentials else ""

        # This changes the actual type of a pydantic class, but its a quick and clean way to format properly
        query.transactionType = ",".join(query.transactionType)  # type: ignore
        url = create_url(4, "insider-trading", api_key, query)

        return get_data_many(url, **kwargs)

    @staticmethod
    def transform_data(data: List[Dict]) -> List[FMPStockInsiderTradingData]:
        """Return the transformed data."""
        return [FMPStockInsiderTradingData(**d) for d in data]



"""FMP Cash Flow Statement Fetcher."""


from datetime import (
    date as dateType,
    datetime,
)
from typing import Any, Dict, List, Optional

from openbb_provider.abstract.fetcher import Fetcher
from openbb_provider.standard_models.cash_flow import (
    CashFlowStatementData,
    CashFlowStatementQueryParams,
)
from pydantic import Field, validator

from openbb_fmp.utils.helpers import create_url, get_data_many


class FMPCashFlowStatementQueryParams(CashFlowStatementQueryParams):
    """FMP Cash Flow Statement Query.

    Source: https://financialmodelingprep.com/developer/docs/#Cash-Flow-Statement
    """

    symbol: str = Field(description="Symbol/CIK of the company.")


class FMPCashFlowStatementData(CashFlowStatementData):
    """FMP Cash Flow Statement Data."""

    class Config:
        """Pydantic alias config using fields dict."""

        fields = {
            "net_cash_flow_from_operating_activities": "netCashProvidedByOperatingActivities",
            "purchases_of_marketable_securities": "purchasesOfInvestments",
            "sales_from_maturities_of_investments": "salesMaturitiesOfInvestments",
            "payments_from_acquisitions": "acquisitionsNet",
            "other_investing_activities": "otherInvestingActivites",
            "net_cash_flow_from_investing_activities": "netCashUsedForInvestingActivites",
            "other_financing_activities": "otherFinancingActivites",
            "net_cash_flow_from_financing_activities": "netCashUsedProvidedByFinancingActivities",
        }

    reported_currency: str = Field(description="Reported currency in the statement.")
    filling_date: dateType = Field(description="Filling date.")
    accepted_date: datetime = Field(description="Accepted date.")
    calendar_year: int = Field(description="Calendar year.")

    change_in_working_capital: Optional[int] = Field(
        description="Change in working capital."
    )
    other_working_capital: Optional[int] = Field(description="Other working capital.")
    common_stock_issued: Optional[int] = Field(description="Common stock issued.")
    effect_of_forex_changes_on_cash: Optional[int] = Field(
        description="Effect of forex changes on cash."
    )

    cash_at_beginning_of_period: Optional[int] = Field(
        description="Cash at beginning of period."
    )
    cash_at_end_of_period: Optional[int] = Field(
        description="Cash, cash equivalents, and restricted cash at end of period"
    )
    operating_cash_flow: Optional[int] = Field(description="Operating cash flow.")
    capital_expenditure: Optional[int] = Field(description="Capital expenditure.")
    free_cash_flow: Optional[int] = Field(description="Free cash flow.")

    link: str = Field(description="Link to the statement.")
    final_link: str = Field(description="Link to the final statement.")

    @validator("filing_date", pre=True, check_fields=False)
    def filing_date_validate(cls, v):  # pylint: disable=no-self-argument
        """Validate the filing date."""
        return datetime.strptime(v, "%Y-%m-%d").date()

    @validator("accepted_date", pre=True, check_fields=False)
    def accepted_date_validate(cls, v):  # pylint: disable=no-self-argument
        """Validate the accepted date."""
        return datetime.strptime(v, "%Y-%m-%d %H:%M:%S")


class FMPCashFlowStatementFetcher(
    Fetcher[
        FMPCashFlowStatementQueryParams,
        List[FMPCashFlowStatementData],
    ]
):
    """Transform the query, extract and transform the data from the FMP endpoints."""

    @staticmethod
    def transform_query(params: Dict[str, Any]) -> FMPCashFlowStatementQueryParams:
        """Transform the query params."""
        return FMPCashFlowStatementQueryParams(**params)

    @staticmethod
    def extract_data(
        query: FMPCashFlowStatementQueryParams,
        credentials: Optional[Dict[str, str]],
        **kwargs: Any,
    ) -> List[Dict]:
        """Return the raw data from the FMP endpoint."""
        api_key = credentials.get("fmp_api_key") if credentials else ""

        url = create_url(
            3, f"cash-flow-statement/{query.symbol}", api_key, query, ["symbol"]
        )

        return get_data_many(url, **kwargs)

    @staticmethod
    def transform_data(data: List[Dict]) -> List[FMPCashFlowStatementData]:
        """Return the transformed data."""
        return [FMPCashFlowStatementData(**d) for d in data]



"""FMP Provider module."""
from openbb_provider.abstract.provider import Provider

from openbb_fmp.models.analyst_estimates import FMPAnalystEstimatesFetcher
from openbb_fmp.models.available_indices import FMPAvailableIndicesFetcher
from openbb_fmp.models.balance_sheet import FMPBalanceSheetFetcher
from openbb_fmp.models.balance_sheet_growth import FMPBalanceSheetGrowthFetcher
from openbb_fmp.models.cash_flow import FMPCashFlowStatementFetcher
from openbb_fmp.models.cash_flow_growth import FMPCashFlowStatementGrowthFetcher
from openbb_fmp.models.company_overview import FMPCompanyOverviewFetcher
from openbb_fmp.models.crypto_historical import FMPCryptoHistoricalFetcher
from openbb_fmp.models.dividend_calendar import FMPDividendCalendarFetcher
from openbb_fmp.models.earnings_calendar import FMPEarningsCalendarFetcher
from openbb_fmp.models.earnings_call_transcript import FMPEarningsCallTranscriptFetcher
from openbb_fmp.models.executive_compensation import FMPExecutiveCompensationFetcher
from openbb_fmp.models.financial_ratios import FMPFinancialRatiosFetcher
from openbb_fmp.models.forex_historical import FMPForexHistoricalFetcher
from openbb_fmp.models.forex_pairs import FMPForexPairsFetcher
from openbb_fmp.models.global_news import FMPGlobalNewsFetcher
from openbb_fmp.models.historical_dividends import FMPHistoricalDividendsFetcher
from openbb_fmp.models.historical_employees import FMPHistoricalEmployeesFetcher
from openbb_fmp.models.historical_stock_splits import FMPHistoricalStockSplitsFetcher
from openbb_fmp.models.income_statement import FMPIncomeStatementFetcher
from openbb_fmp.models.income_statement_growth import FMPIncomeStatementGrowthFetcher
from openbb_fmp.models.institutional_ownership import FMPInstitutionalOwnershipFetcher
from openbb_fmp.models.key_executives import FMPKeyExecutivesFetcher
from openbb_fmp.models.key_metrics import FMPKeyMetricsFetcher
from openbb_fmp.models.major_indices_constituents import (
    FMPMajorIndicesConstituentsFetcher,
)
from openbb_fmp.models.major_indices_historical import FMPMajorIndicesHistoricalFetcher
from openbb_fmp.models.price_target import FMPPriceTargetFetcher
from openbb_fmp.models.price_target_consensus import FMPPriceTargetConsensusFetcher
from openbb_fmp.models.revenue_business_line import FMPRevenueBusinessLineFetcher
from openbb_fmp.models.revenue_geographic import FMPRevenueGeographicFetcher
from openbb_fmp.models.risk_premium import FMPRiskPremiumFetcher
from openbb_fmp.models.sec_filings import FMPSECFilingsFetcher
from openbb_fmp.models.share_statistics import FMPShareStatisticsFetcher
from openbb_fmp.models.stock_historical import FMPStockHistoricalFetcher
from openbb_fmp.models.stock_insider_trading import FMPStockInsiderTradingFetcher
from openbb_fmp.models.stock_multiples import FMPStockMultiplesFetcher
from openbb_fmp.models.stock_news import FMPStockNewsFetcher
from openbb_fmp.models.stock_ownership import FMPStockOwnershipFetcher
from openbb_fmp.models.stock_peers import FMPStockPeersFetcher
from openbb_fmp.models.stock_quote import FMPStockQuoteFetcher
from openbb_fmp.models.stock_splits import FMPStockSplitCalendarFetcher
from openbb_fmp.models.treasury_rates import FMPTreasuryRatesFetcher

fmp_provider = Provider(
    name="fmp",
    website="https://financialmodelingprep.com/",
    description="""Financial Modeling Prep is a new concept that informs you about
    stock market information (news, currencies, and stock prices).""",
    required_credentials=["api_key"],
    fetcher_dict={
        "KeyExecutives": FMPKeyExecutivesFetcher,
        "StockHistorical": FMPStockHistoricalFetcher,
        "GlobalNews": FMPGlobalNewsFetcher,
        "StockNews": FMPStockNewsFetcher,
        "IncomeStatement": FMPIncomeStatementFetcher,
        "IncomeStatementGrowth": FMPIncomeStatementGrowthFetcher,
        "BalanceSheet": FMPBalanceSheetFetcher,
        "BalanceSheetGrowth": FMPBalanceSheetGrowthFetcher,
        "CashFlowStatement": FMPCashFlowStatementFetcher,
        "CashFlowStatementGrowth": FMPCashFlowStatementGrowthFetcher,
        "ShareStatistics": FMPShareStatisticsFetcher,
        "MajorIndicesHistorical": FMPMajorIndicesHistoricalFetcher,
        "RevenueGeographic": FMPRevenueGeographicFetcher,
        "RevenueBusinessLine": FMPRevenueBusinessLineFetcher,
        "InstitutionalOwnership": FMPInstitutionalOwnershipFetcher,
        "CompanyOverview": FMPCompanyOverviewFetcher,
        "StockInsiderTrading": FMPStockInsiderTradingFetcher,
        "StockOwnership": FMPStockOwnershipFetcher,
        "PriceTargetConsensus": FMPPriceTargetConsensusFetcher,
        "PriceTarget": FMPPriceTargetFetcher,
        "AnalystEstimates": FMPAnalystEstimatesFetcher,
        "EarningsCalendar": FMPEarningsCalendarFetcher,
        "EarningsCallTranscript": FMPEarningsCallTranscriptFetcher,
        "HistoricalStockSplits": FMPHistoricalStockSplitsFetcher,
        "StockSplitCalendar": FMPStockSplitCalendarFetcher,
        "HistoricalDividends": FMPHistoricalDividendsFetcher,
        "KeyMetrics": FMPKeyMetricsFetcher,
        "SECFilings": FMPSECFilingsFetcher,
        "TreasuryRates": FMPTreasuryRatesFetcher,
        "ExecutiveCompensation": FMPExecutiveCompensationFetcher,
        "CryptoHistorical": FMPCryptoHistoricalFetcher,
        "ForexHistorical": FMPForexHistoricalFetcher,
        "ForexPairs": FMPForexPairsFetcher,
        "StockPeers": FMPStockPeersFetcher,
        "StockMultiples": FMPStockMultiplesFetcher,
        "HistoricalEmployees": FMPHistoricalEmployeesFetcher,
        "AvailableIndices": FMPAvailableIndicesFetcher,
        "RiskPremium": FMPRiskPremiumFetcher,
        "MajorIndicesConstituents": FMPMajorIndicesConstituentsFetcher,
        "DividendCalendar": FMPDividendCalendarFetcher,
        "StockQuote": FMPStockQuoteFetcher,
        "FinancialRatios": FMPFinancialRatiosFetcher,
    },
)



"""FMP Stocks end of day fetcher."""

from datetime import datetime
from typing import Any, Dict, List, Literal, Optional

from dateutil.relativedelta import relativedelta
from openbb_provider.abstract.fetcher import Fetcher
from openbb_provider.standard_models.stock_eod import StockEODData, StockEODQueryParams
from openbb_provider.utils.helpers import get_querystring
from pydantic import Field, NonNegativeInt, validator

from openbb_fmp.utils.helpers import get_data_many


class FMPStockEODQueryParams(StockEODQueryParams):
    """FMP Stock end of day Query.

    Source: https://financialmodelingprep.com/developer/docs/#Stock-Historical-Price
    """

    timeseries: Optional[NonNegativeInt] = Field(
        default=None, description="Number of days to look back."
    )
    interval: Literal[
        "1min", "5min", "15min", "30min", "1hour", "4hour", "1day"
    ] = Field(default="1day", description="Interval of the data to fetch.")


class FMPStockEODData(StockEODData):
    """FMP Stock end of day Data."""

    adjClose: Optional[float] = Field(description="Adjusted Close Price of the symbol.")
    unadjustedVolume: Optional[float] = Field(
        description="Unadjusted volume of the symbol."
    )
    change: Optional[float] = Field(
        description="Change in the price of the symbol from the previous day."
    )
    changePercent: Optional[float] = Field(
        description=r"Change \% in the price of the symbol."
    )
    vwap: Optional[float] = Field(
        description="Volume Weighted Average Price of the symbol."
    )
    label: Optional[str] = Field(description="Human readable format of the date.")
    changeOverTime: Optional[float] = Field(
        description=r"Change \% in the price of the symbol over a period of time."
    )

    @validator("date", pre=True, check_fields=False)
    def date_validate(cls, v):  # pylint: disable=E0213
        """Return the date as a datetime object."""
        try:
            return datetime.strptime(v, "%Y-%m-%d %H:%M:%S")
        except ValueError:
            return datetime.strptime(v, "%Y-%m-%d")


class FMPStockEODFetcher(
    Fetcher[
        FMPStockEODQueryParams,
        List[FMPStockEODData],
    ]
):
    """Transform the query, extract and transform the data from the FMP endpoints."""

    @staticmethod
    def transform_query(params: Dict[str, Any]) -> FMPStockEODQueryParams:
        """Transform the query params."""
        transformed_params = params

        now = datetime.now().date()
        if params.get("start_date") is None:
            transformed_params["start_date"] = now - relativedelta(years=1)

        if params.get("end_date") is None:
            transformed_params["end_date"] = now

        return FMPStockEODQueryParams(**transformed_params)

    @staticmethod
    def extract_data(
        query: FMPStockEODQueryParams,
        credentials: Optional[Dict[str, str]],
        **kwargs: Any,
    ) -> List[Dict]:
        """Return the raw data from the FMP endpoint."""
        api_key = credentials.get("fmp_api_key") if credentials else ""

        base_url = "https://financialmodelingprep.com/api/v3"
        query_str = (
            get_querystring(query.dict(), ["symbol"])
            .replace("start_date", "from")
            .replace("end_date", "to")
        )

        url_params = f"{query.symbol}?{query_str}&apikey={api_key}"
        url = f"{base_url}/historical-chart/{query.interval}/{url_params}"

        if query.interval == "1day":
            url = f"{base_url}/historical-price-full/{url_params}"

        return get_data_many(url, "historical", **kwargs)

    @staticmethod
    def transform_data(data: List[Dict]) -> List[FMPStockEODData]:
        """Return the transformed data."""
        return [FMPStockEODData(**d) for d in data]



"""FMP Cryptos end of day fetcher."""


from datetime import datetime
from typing import Any, Dict, List, Optional

from dateutil.relativedelta import relativedelta
from openbb_provider.abstract.fetcher import Fetcher
from openbb_provider.standard_models.crypto_eod import (
    CryptoEODData,
    CryptoEODQueryParams,
)
from openbb_provider.utils.helpers import get_querystring
from pydantic import Field, NonNegativeInt, validator

from openbb_fmp.utils.helpers import get_data_many


class FMPCryptoEODQueryParams(CryptoEODQueryParams):
    # noqa: E501
    """FMP Crypto end of day Query.

    Source:
    https://site.financialmodelingprep.com/developer/docs/cryptocurrency-historical-data-api/#Historical-Daily-Prices
    """

    timeseries: Optional[NonNegativeInt] = Field(
        default=None, description="Number of days to look back."
    )


class FMPCryptoEODData(CryptoEODData):
    """FMP Crypto end of day Data."""

    adjClose: float = Field(
        description="Adjusted Close Price of the symbol.", alias="adj_close"
    )
    unadjustedVolume: float = Field(
        description="Unadjusted volume of the symbol.", alias="unadjusted_volume"
    )
    change: float = Field(
        description="Change in the price of the symbol from the previous day.",
        alias="change",
    )
    changePercent: float = Field(
        description=r"Change \% in the price of the symbol.", alias="change_percent"
    )
    label: str = Field(description="Human readable format of the date.")
    changeOverTime: float = Field(
        description=r"Change \% in the price of the symbol over a period of time.",
        alias="change_over_time",
    )

    @validator("date", pre=True)
    def date_validate(cls, v):  # pylint: disable=E0213
        """Return the date as a datetime object."""
        return datetime.strptime(v, "%Y-%m-%d")


class FMPCryptoEODFetcher(
    Fetcher[
        FMPCryptoEODQueryParams,
        List[FMPCryptoEODData],
    ]
):
    """Transform the query, extract and transform the data from the FMP endpoints."""

    @staticmethod
    def transform_query(params: Dict[str, Any]) -> FMPCryptoEODQueryParams:
        """Transform the query params. Start and end dates are set to 1 year interval."""
        transformed_params = params

        now = datetime.now().date()
        if params.get("start_date") is None:
            transformed_params["start_date"] = now - relativedelta(years=1)

        if params.get("end_date") is None:
            transformed_params["end_date"] = now
        return FMPCryptoEODQueryParams(**transformed_params)

    @staticmethod
    def extract_data(
        query: FMPCryptoEODQueryParams,
        credentials: Optional[Dict[str, str]],
        **kwargs: Any,
    ) -> List[Dict]:
        """Return the raw data from the FMP endpoint."""
        api_key = credentials.get("fmp_api_key") if credentials else ""

        base_url = "https://financialmodelingprep.com/api/v3/"
        query_str = get_querystring(query.dict(by_alias=True), ["symbol"])
        query_str = query_str.replace("start_date", "from").replace("end_date", "to")
        url = f"{base_url}historical-price-full/crypto/{query.symbol}?{query_str}&apikey={api_key}"

        return get_data_many(url, "historical", **kwargs)

    @staticmethod
    def transform_data(data: List[Dict]) -> List[FMPCryptoEODData]:
        """Return the transformed data."""
        return [FMPCryptoEODData(**d) for d in data]



"""FMP Forex available pairs fetcher."""


from typing import Any, Dict, List, Optional

from openbb_provider.abstract.fetcher import Fetcher
from openbb_provider.standard_models.forex_pairs import (
    ForexPairsData,
    ForexPairsQueryParams,
)
from pydantic import Field

from openbb_fmp.utils.helpers import get_data_many


class FMPForexPairsQueryParams(ForexPairsQueryParams):
    """FMP Forex available pairs Query.

    Source: https://site.financialmodelingprep.com/developer/docs/#Historical-Forex-Price
    """


class FMPForexPairsData(ForexPairsData):
    """FMP Forex available pairs Data."""

    symbol: str = Field(description="Symbol of the currency pair.")
    currency: str = Field(description="Base currency of the currency pair.")
    stockExchange: Optional[str] = Field(
        description="Stock exchange of the currency pair.",
        alias="stock_exchange",
    )
    exchange_short_name: Optional[str] = Field(
        description="Short name of the stock exchange of the currency pair.",
        alias="exchange_short_name",
    )


class FMPForexPairsFetcher(
    Fetcher[
        FMPForexPairsQueryParams,
        List[FMPForexPairsData],
    ]
):
    """Transform the query, extract and transform the data from the FMP endpoints."""

    @staticmethod
    def transform_query(params: Dict[str, Any]) -> FMPForexPairsQueryParams:
        """Transform the query params."""
        return FMPForexPairsQueryParams(**params)

    @staticmethod
    def extract_data(
        query: FMPForexPairsQueryParams,
        credentials: Optional[Dict[str, str]],
        **kwargs: Any,
    ) -> List[Dict]:
        """Return the raw data from the FMP endpoint."""
        api_key = credentials.get("fmp_api_key") if credentials else ""

        base_url = "https://financialmodelingprep.com/api/v3"
        url = f"{base_url}/symbol/available-forex-currency-pairs?apikey={api_key}"

        return get_data_many(url, **kwargs)

    @staticmethod
    def transform_data(data: List[Dict]) -> List[FMPForexPairsData]:
        """Return the transformed data."""
        return [FMPForexPairsData(**d) for d in data]



"""FMP Earnings Calendar fetcher."""


from datetime import datetime
from typing import Any, Dict, List, Optional

from openbb_provider.abstract.fetcher import Fetcher
from openbb_provider.standard_models.earnings_calendar import (
    EarningsCalendarData,
    EarningsCalendarQueryParams,
)
from pydantic import validator

from openbb_fmp.utils.helpers import create_url, get_data_many


class FMPEarningsCalendarQueryParams(EarningsCalendarQueryParams):
    """FMP Earnings Calendar Query.

    Source: https://site.financialmodelingprep.com/developer/docs/earnings-calendar-api/
    """


class FMPEarningsCalendarData(EarningsCalendarData):
    """FMP Earnings Calendar Data."""

    @validator("date", pre=True, check_fields=False)
    def date_validate(cls, v: str):  # pylint: disable=E0213
        """Return the date as a datetime object."""
        return datetime.strptime(v, "%Y-%m-%d") if v else None

    @validator("updatedFromDate", pre=True, check_fields=False)
    def updated_from_date_validate(cls, v: str):  # pylint: disable=E0213
        """Return the updated from date as a datetime object."""
        return datetime.strptime(v, "%Y-%m-%d") if v else None

    @validator("fiscalDateEnding", pre=True, check_fields=False)
    def fiscal_date_ending_validate(cls, v: str):  # pylint: disable=E0213
        """Return the fiscal date ending as a datetime object."""
        return datetime.strptime(v, "%Y-%m-%d") if v else None


class FMPEarningsCalendarFetcher(
    Fetcher[
        FMPEarningsCalendarQueryParams,
        List[FMPEarningsCalendarData],
    ]
):
    """Transform the query, extract and transform the data from the FMP endpoints."""

    @staticmethod
    def transform_query(params: Dict[str, Any]) -> FMPEarningsCalendarQueryParams:
        """Transform the query params."""
        return FMPEarningsCalendarQueryParams(**params)

    @staticmethod
    def extract_data(
        query: FMPEarningsCalendarQueryParams,
        credentials: Optional[Dict[str, str]],
        **kwargs: Any,
    ) -> List[Dict]:
        """Return the raw data from the FMP endpoint."""
        api_key = credentials.get("fmp_api_key") if credentials else ""

        url = create_url(
            3, f"historical/earning_calendar/{query.symbol}", api_key, query, ["symbol"]
        )

        return get_data_many(url, **kwargs)

    @staticmethod
    def transform_data(data: List[Dict]) -> List[FMPEarningsCalendarData]:
        """Return the transformed data."""
        return [FMPEarningsCalendarData(**d) for d in data]



"""FMP Financial Ratios Fetcher."""


from concurrent.futures import ThreadPoolExecutor
from itertools import repeat
from typing import Any, Dict, List, Literal, Optional

from openbb_provider.abstract.fetcher import Fetcher
from openbb_provider.standard_models.base import FinancialStatementQueryParams
from openbb_provider.standard_models.financial_ratios import FinancialRatiosData

from openbb_fmp.utils.helpers import get_data_many

PeriodType = Literal["annual", "quarter"]


class FMPFinancialRatiosQueryParams(FinancialStatementQueryParams):
    """FMP Financial Ratios QueryParams.

    Source: https://financialmodelingprep.com/developer/docs/#Company-Financial-Ratios

    Symbol must be provided.
    """


class FMPFinancialRatiosData(FinancialRatiosData):
    """FMP Financial Ratios Data."""


class FMPFinancialRatiosFetcher(
    Fetcher[
        FMPFinancialRatiosQueryParams,
        List[FMPFinancialRatiosData],
    ]
):
    """Transform the query, extract and transform the data from the FMP endpoints."""

    @staticmethod
    def transform_query(params: Dict[str, Any]) -> FMPFinancialRatiosQueryParams:
        """Transform the query params."""
        return FMPFinancialRatiosQueryParams(**params)

    @staticmethod
    def extract_data(
        query: FMPFinancialRatiosQueryParams,
        credentials: Optional[Dict[str, str]],
        **kwargs: Any,
    ) -> List[Dict]:
        """Return the raw data from the FMP endpoint."""
        api_key = credentials.get("fmp_api_key") if credentials else ""

        base_url = "https://financialmodelingprep.com/api/v3"
        data: List[Dict] = []

        def multiple_symbols(symbol: str, data: List[Dict]) -> None:
            url = (
                f"{base_url}/ratios/{symbol}?"
                f"period={query.period}&limit={query.limit}&apikey={api_key}"
            )
            return data.extend(get_data_many(url, **kwargs))

        with ThreadPoolExecutor() as executor:
            executor.map(multiple_symbols, query.symbol.split(","), repeat(data))

        return data

    @staticmethod
    def transform_data(data: List[Dict]) -> List[FMPFinancialRatiosData]:
        """Return the transformed data."""
        return [FMPFinancialRatiosData(**d) for d in data]
